@@ -23,21 +23,21 @@ sub demand
 
     if (my $prefetch = $storage->{PREFETCH}{$class}{$member}{$storage->id($obj)})
     {
-	%coll = %$prefetch;
+		%coll = %$prefetch;
     }
     else
     {
-	my $cursor = $self->cursor($def, $storage, $obj, $member);
+		my $cursor = $self->cursor($def, $storage, $obj, $member);
 
-	for (my $item = $cursor->select; $item; $item = $cursor->next)
-	{
-	    my $slot = shift @{ $cursor->{-residue} };
-	    $coll{$slot} = $item;
-	}
+		for (my $item = $cursor->select; $item; $item = $cursor->next)
+		{
+			my $slot = shift @{ $cursor->{-residue} };
+			$coll{$slot} = $item;
+		}
     }
 
     $storage->{scratch}{ref($self)}{$storage->id($obj)}{$member} = {
-	map { $_ => ($coll{$_} && $storage->id( $coll{$_} ) ) } keys %coll };
+																	map { $_ => ($coll{$_} && $storage->id( $coll{$_} ) ) } keys %coll };
 
     return \%coll;
 }
@@ -48,12 +48,17 @@ sub save
 
     foreach my $coll (keys %$members)
     {
-	next if tied $obj->{$coll};
+		next if tied $obj->{$coll};
+		next unless defined $obj->{$coll};
 
-	foreach my $item (values %{ $obj->{$coll} } )
-	{
-	    $storage->insert($item) unless $storage->id($item);
-	}
+		my $class = $members->{$coll}{class};
+
+		foreach my $item (values %{ $obj->{$coll} } )
+		{
+			Tangram::Coll::bad_type($obj, $coll, $class, $item)
+					if $^W && !$item->isa($class);
+			$storage->insert($item) unless $storage->id($item);
+		}
     }
 
     $storage->defer(sub { $self->defered_save(shift, $obj, $members, $id) } );
